@@ -15,7 +15,7 @@ use symphonia_core::codecs::{
     CODEC_TYPE_PCM_F64LE, CODEC_TYPE_PCM_MULAW, CODEC_TYPE_PCM_S16LE, CODEC_TYPE_PCM_S24LE,
     CODEC_TYPE_PCM_S32LE, CODEC_TYPE_PCM_U8,
 };
-use symphonia_core::errors::{decode_error, unsupported_error, Result};
+use symphonia_core::errors::{decode_error, unsupported_error, Error, Result};
 use symphonia_core::io::ReadBytes;
 use symphonia_core::meta::Tag;
 use symphonia_metadata::riff;
@@ -38,6 +38,63 @@ impl ParseChunkTag for NullChunks {
     }
 }
 
+fn fix_channel_mask(mut channel_mask: u32, n_channels: u16) -> u32 {
+    let channel_diff = n_channels as i32 - channel_mask.count_ones() as i32;
+
+    if channel_diff != 0 {
+        info!("Channel mask not set correctly, channel positions may be incorrect!");
+    }
+
+    // Check that the number of ones in the channel mask match the number of channels.
+    if channel_diff > 0 {
+        // Too few ones in mask so add extra ones above the most significant one
+        let shift = 32 - (!channel_mask).leading_ones();
+        channel_mask |= ((1 << channel_diff) - 1) << shift;
+    }
+    else {
+        // Too many ones in mask so remove the most significant extra ones
+        while channel_mask.count_ones() != n_channels as u32 {
+            let highest_one = 31 - (!channel_mask).leading_ones();
+            channel_mask &= !(1 << highest_one);
+        }
+    }
+
+    channel_mask
+}
+
+#[test]
+fn test_fix_channel_mask() {
+    // Too few
+    assert_eq!(fix_channel_mask(0, 9), 0b111111111);
+    assert_eq!(fix_channel_mask(0b101000, 5), 0b111101000);
+
+    // Too many
+    assert_eq!(fix_channel_mask(0b1111111, 0), 0);
+    assert_eq!(fix_channel_mask(0b101110111010, 5), 0b10111010);
+    assert_eq!(fix_channel_mask(0xFFFFFFFF, 8), 0b11111111);
+}
+
+fn try_channel_count_to_mask(count: u16) -> Result<Channels> {
+    (1..=32)
+        .contains(&count)
+        .then(|| Channels::from_bits(((1u64 << count) - 1) as u32))
+        .flatten()
+        .ok_or(Error::DecodeError("wav: invalid channel count"))
+}
+
+#[test]
+fn test_try_channel_count_to_mask() {
+    assert!(try_channel_count_to_mask(0).is_err());
+
+    for i in 1..27 {
+        assert!(try_channel_count_to_mask(i).is_ok());
+    }
+
+    for i in 27..u16::MAX {
+        assert!(try_channel_count_to_mask(i).is_err());
+    }
+}
+
 /// `ChunksReader` reads chunks from a `ByteStream`. It is generic across a type, usually an enum,
 /// implementing the `ParseChunkTag` trait. When a new chunk is encountered in the stream,
 /// `parse_tag` on T is called to return an object capable of parsing/reading that chunk or `None`.
@@ -269,14 +326,7 @@ impl WaveFormatChunk {
             }
         };
 
-        // The PCM format only supports 1 or 2 channels, for mono and stereo channel layouts,
-        // respectively.
-        let channels = match n_channels {
-            1 => Channels::FRONT_LEFT,
-            2 => Channels::FRONT_LEFT | Channels::FRONT_RIGHT,
-            _ => return decode_error("wav: channel layout is not stereo or mono for fmt_pcm"),
-        };
-
+        let channels = try_channel_count_to_mask(n_channels)?;
         Ok(WaveFormatData::Pcm(WaveFormatPcm { bits_per_sample, channels, codec }))
     }
 
@@ -309,13 +359,7 @@ impl WaveFormatChunk {
         }
         reader.ignore_bytes(extra_size)?;
 
-        // The ADPCM format only supports 1 or 2 channels, for mono and stereo channel layouts,
-        // respectively.
-        let channels = match n_channels {
-            1 => Channels::FRONT_LEFT,
-            2 => Channels::FRONT_LEFT | Channels::FRONT_RIGHT,
-            _ => return decode_error("wav: channel layout is not stereo or mono for fmt_adpcm"),
-        };
+        let channels = try_channel_count_to_mask(n_channels)?;
         Ok(WaveFormatData::Adpcm(WaveFormatAdpcm { bits_per_sample, channels, codec }))
     }
 
@@ -327,15 +371,21 @@ impl WaveFormatChunk {
     ) -> Result<WaveFormatData> {
         // WaveFormat for a IEEE format should not be extended, but it may still have an extra data
         // length parameter.
-        if len == 18 {
-            let extra_size = reader.read_u16()?;
-
-            if extra_size != 0 {
-                return decode_error("wav: extra data not expected for fmt_ieee chunk");
+        match len {
+            16 => (),
+            18 => {
+                let extra_size = reader.read_u16()?;
+                if extra_size != 0 {
+                    return decode_error("wav: extra data not expected for fmt_ieee chunk");
+                }
             }
-        }
-        else if len > 16 {
-            return decode_error("wav: malformed fmt_ieee chunk");
+            40 => {
+                // WAVEFORMATEXTENSIBLE is used for formats having more than two channels
+                // or higher sample resolutions than allowed by WAVEFORMATEX but for now
+                // we just ignore it
+                let _ = reader.ignore_bytes(40 - 16);
+            }
+            _ => return decode_error("wav: malformed fmt_ieee chunk"),
         }
 
         // Officially, only 32-bit floats are supported, but Symphonia can handle 64-bit floats.
@@ -348,14 +398,7 @@ impl WaveFormatChunk {
             _ => return decode_error("wav: bits per sample for fmt_ieee must be 32 or 64 bits"),
         };
 
-        // The IEEE format only supports 1 or 2 channels, for mono and stereo channel layouts,
-        // respectively.
-        let channels = match n_channels {
-            1 => Channels::FRONT_LEFT,
-            2 => Channels::FRONT_LEFT | Channels::FRONT_RIGHT,
-            _ => return decode_error("wav: channel layout is not stereo or mono for fmt_ieee"),
-        };
-
+        let channels = try_channel_count_to_mask(n_channels)?;
         Ok(WaveFormatData::IeeeFloat(WaveFormatIeeeFloat { channels, codec }))
     }
 
@@ -394,12 +437,7 @@ impl WaveFormatChunk {
             );
         }
 
-        let channel_mask = reader.read_u32()?;
-
-        // The number of ones in the channel mask should match the number of channels.
-        if channel_mask.count_ones() != u32::from(n_channels) {
-            return decode_error("wav: channel mask mismatch with number of channels for fmt_ext");
-        }
+        let channel_mask = fix_channel_mask(reader.read_u32()?, n_channels);
 
         // Try to map channels.
         let channels = match Channels::from_bits(channel_mask) {
@@ -506,12 +544,7 @@ impl WaveFormatChunk {
             reader.ignore_bytes(u64::from(extra_size))?;
         }
 
-        let channels = match n_channels {
-            1 => Channels::FRONT_LEFT,
-            2 => Channels::FRONT_LEFT | Channels::FRONT_RIGHT,
-            _ => return decode_error("wav: channel layout is not stereo or mono for fmt_alaw"),
-        };
-
+        let channels = try_channel_count_to_mask(n_channels)?;
         Ok(WaveFormatData::ALaw(WaveFormatALaw { codec: CODEC_TYPE_PCM_ALAW, channels }))
     }
 
@@ -530,12 +563,7 @@ impl WaveFormatChunk {
             reader.ignore_bytes(u64::from(extra_size))?;
         }
 
-        let channels = match n_channels {
-            1 => Channels::FRONT_LEFT,
-            2 => Channels::FRONT_LEFT | Channels::FRONT_RIGHT,
-            _ => return decode_error("wav: channel layout is not stereo or mono for fmt_mulaw"),
-        };
-
+        let channels = try_channel_count_to_mask(n_channels)?;
         Ok(WaveFormatData::MuLaw(WaveFormatMuLaw { codec: CODEC_TYPE_PCM_MULAW, channels }))
     }
 
